﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CatanGame.Model
{
    public class CatanBoard
    {
        #region Private fields
        private readonly static Dictionary<ResourceType, int> resourcePossibilities = new Dictionary<ResourceType, int> {
            { ResourceType.None, 1 },
            { ResourceType.Grain, 4 },
            { ResourceType.Lumber, 4 },
            { ResourceType.Wool, 4 },
            { ResourceType.Brick, 3 },
            { ResourceType.Ore, 3 },
        };

        private Player[] _players;
        private Node[,,] _nodes;
        private List<(NodeCoordinates NodeCoords, byte PlayerId)>[,,] _roads;
        private Field[,] _fields;
        private List<Field>[] _fieldNumbers;
        private int _currentPlayer;
        private static Random rand = new Random();
        private bool _initialTurn;
        private int _turnIncrement;
        #endregion

        #region Public properties
        public Player[] Players { get => _players; }
        public Node[,,] Nodes { get => _nodes; }
        public List<(NodeCoordinates NodeCoords, byte playerId)>[,,] Roads { get => _roads; }
        public Field[,] Fields { get => _fields; }
        public (int, int) RobberPosition { get; set; }
        public Player LongestRoadOwner { get; set; }
        public Player LargestArmyOwner { get; set; }
        public bool IsGameOver { get; set; }
        public int Size { get; set; }

        public bool InitialTurn { get => _initialTurn; }

        public int CurrentPlayerId { get => _currentPlayer; }

        public int HexagonCount { get => Size * Size - (Size / 2 + 1) * Size / 2; }

        public int VertexCount { get => HexagonCount * 2 + Size * 3 + 1; }
        #endregion


        #region Public constructors
        public CatanBoard(List<String> playerNames, int size = 7)
        {
            if (size % 2 == 0)
                throw new InvalidOperationException("size can't be an even number!");
            Size = size;
 
            InitializeFields();
            InitializeNodes();
            InitializeRoads();
            InitializePlayers(playerNames);

        }
        #endregion

        #region Public methods

        public void BuyCard(Player p)
        {
            if (p.CanBuyDevelopmentCard)
            {
                p.DevelopmentCards[(CardType)rand.Next(0, 2)] += 1;
                p.DevelopmentCardBought();
            }
        }

        public void PlaceRoad((NodeCoordinates FirstNode, NodeCoordinates SecondNode ) coords, Player p)
        {
            if (CanPlaceRoad(coords, p))
            {
                _roads[coords.FirstNode.R, coords.FirstNode.Q, (int) coords.FirstNode.Dir].Add((coords.SecondNode, p.Id));
                _roads[coords.SecondNode.R, coords.SecondNode.Q, (int)coords.SecondNode.Dir].Add((coords.FirstNode, p.Id));
                p.RoadBuilt();
            }
            if (_initialTurn)
                StepGame();
        }

        public List<(NodeCoordinates, NodeCoordinates)> EnabledRoads(Player p)
        {
            List<(NodeCoordinates, NodeCoordinates)> answer = new();
            for (int r = 0; r < Size; r++)
                for (int q = 0; q < Size; q++)
                    for (int d = 0; d < 2; d++)
                        foreach ((NodeCoordinates, NodeCoordinates) road in EnabledRoadsFromNode(p, new NodeCoordinates(r, q, (Direction) d)))
                            answer.Add(road);
                    
            return answer;
        }

        public List<(NodeCoordinates, NodeCoordinates)> EnabledRoadsFromNode(Player p, NodeCoordinates current)
        {
            List<(NodeCoordinates, NodeCoordinates)> answer = new List<(NodeCoordinates, NodeCoordinates)>();
            List<NodeCoordinates> roads = GetNeighborCoords(current);
            foreach (NodeCoordinates road in roads)
            {
                if (current.R >= 0 && current.Q >= 0 && road.R >= 0 && road.Q >= 0 && current.R < 7 && current.Q < 7 && road.R < 7 && road.Q < 7)

                    // the nodes (Not the roads) should not be null
                    if (_nodes[current.R, current.Q, (int)current.Dir] is not null && _nodes[road.R, road.Q, (int)road.Dir] is not null)
                    
                        if (!IsRoad(current, road) && CanPlaceRoad((current, road), p))
                            answer.Add((current, road));
                    
            }
            return answer;
        }

        public List<(NodeCoordinates, byte)> RoadsFrom(NodeCoordinates coords)
        {
            return _roads[coords.R, coords.Q, (int)coords.Dir];
        }

        public bool IsRoad(NodeCoordinates FirstCoords, NodeCoordinates SecondCoords)
        {
            bool answer = false;
            List<(NodeCoordinates coords, byte Id)> roads = RoadsFrom(FirstCoords);
            foreach ((NodeCoordinates coords, byte Id)road in roads)
            {
                if (road.coords.Equals(SecondCoords))
                    answer = true;
            }

            return answer;
        }

        public bool PlaceSettlement(NodeCoordinates node, Player p)
        {
            if (CanPlaceSettlement(node, p))
            {
                House building = _nodes[node.R, node.Q, (int)node.Dir].Building;
                building.Type = HouseType.Settlement;
                building.Owner = p;

                p.AddedSettlement();
                return true;
            }
            return false;
        }

        public bool UpgradeToCity(NodeCoordinates node, Player p)
        {
            if (CanUpgradeToCity(node, p))
            {
                _nodes[node.R, node.Q, (int)node.Dir].Building.Type = HouseType.City;
                p.UpgradedToCity();
                return true;
            }
            return false;
        }

        public void StepGame()
        {
            _currentPlayer += _turnIncrement;
            if (_currentPlayer == _players.Length)
            {
                if (_initialTurn)
                {
                    _turnIncrement *= -1;
                    _currentPlayer += _turnIncrement;
                }
                else
                    _currentPlayer = 0;
            } 
            if (_currentPlayer == -1)
            {
                _initialTurn = false;
                _turnIncrement *= -1;
                _currentPlayer += _turnIncrement;
            }
        }

        public void MoveRobber((int r, int q) coords)
        {
            RobberPosition = (coords.r, coords.q);
        }

        public (int, int) Roll()
        {
            return (rand.Next(1, 7), rand.Next(1, 7));
        }

        public void SendResources(byte id)
        {
            for (int r = 0; r < Size; r++)
                for (int q = 0; q < Size; q++)
                    if (_fields[r,q].FieldID == id && (r, q) != RobberPosition)
                        _fields[r, q].SendResources();
        }

        public bool CanPlaceRoad((NodeCoordinates FirstNode, NodeCoordinates SecondNode) coords, Player p)
        {
            List<NodeCoordinates> firstNodeNeighbors = GetNeighborCoords(coords.FirstNode);
            List<NodeCoordinates> secondNodeNeighbors = GetNeighborCoords(coords.SecondNode);
            bool answer = false;

            if (_initialTurn)
            {
                if (Nodes[coords.FirstNode.R, coords.FirstNode.Q, (int)coords.FirstNode.Dir] is not null && Nodes[coords.FirstNode.R, coords.FirstNode.Q, (int)coords.FirstNode.Dir].Building.Owner is not null)
                    answer |= Nodes[coords.FirstNode.R, coords.FirstNode.Q, (int)coords.FirstNode.Dir].Building.Owner.Id == _currentPlayer;
                if (Nodes[coords.SecondNode.R, coords.SecondNode.Q, (int)coords.SecondNode.Dir] is not null && Nodes[coords.SecondNode.R, coords.SecondNode.Q, (int)coords.SecondNode.Dir].Building.Owner is not null)
                    answer |= Nodes[coords.SecondNode.R, coords.SecondNode.Q, (int)coords.SecondNode.Dir].Building.Owner.Id == _currentPlayer;
            }

            foreach ((NodeCoordinates NodeCoords, byte PlayerId) Road in Roads[coords.FirstNode.R, coords.FirstNode.Q, (int)coords.FirstNode.Dir])
                answer |= IsRoad(Road.NodeCoords, coords.FirstNode) && Road.PlayerId == _currentPlayer && 
                          (Nodes[coords.FirstNode.R, coords.FirstNode.Q, (int)coords.FirstNode.Dir].Building.Type == HouseType.NoBuilding || Nodes[coords.FirstNode.R, coords.FirstNode.Q, (int)coords.FirstNode.Dir].Building.Owner.Id == _currentPlayer);

            foreach ((NodeCoordinates NodeCoords, byte PlayerId) Road in Roads[coords.SecondNode.R, coords.SecondNode.Q, (int)coords.SecondNode.Dir])
                answer |= IsRoad(Road.NodeCoords, coords.SecondNode) && Road.PlayerId == _currentPlayer && 
                          (Nodes[coords.SecondNode.R, coords.SecondNode.Q, (int)coords.SecondNode.Dir].Building.Type == HouseType.NoBuilding || Nodes[coords.SecondNode.R, coords.SecondNode.Q, (int)coords.SecondNode.Dir].Building.Owner.Id == _currentPlayer);

            return answer;
        }

        public bool CanPlaceSettlement(NodeCoordinates node, Player p)
        {
            if (_currentPlayer != p.Id)
                return false;

            List<NodeCoordinates> neighbors = GetNeighborCoords(node);
            bool answer = Nodes[node.R, node.Q, (int)node.Dir].Building.Type == HouseType.NoBuilding 
                && (p.CanBuildSettlement || _initialTurn);

            foreach (NodeCoordinates NodeCoords in neighbors)
            {
                Node neighNode = Nodes[NodeCoords.R, NodeCoords.Q, (int)NodeCoords.Dir];
                answer &= neighNode is null || neighNode.Building.Type == HouseType.NoBuilding;
            }
            return answer;
        }

        public bool CanUpgradeToCity(NodeCoordinates node, Player p)
        {
            return Nodes[node.R, node.Q, (int)node.Dir].Building.Owner == p && p.CanBuildCity;
        }

        #endregion

        #region Private methods
        private void InitializeFields()
        {
            _fields = new Field[Size, Size];
            Dictionary<ResourceType, int> possibleResources = new(resourcePossibilities);

            int innerBoardSize = Size - 2;
            for(int q = 0; q < Size; ++q)
            {
                for (int r = 0; r < Size; ++r)
                {
                    if (Math.Abs(q + r - Size + 1) < 1 + Size / 2) 
                    {
                        _fields[r,q] = new Field();

                        if (IsInnerBoard(r, q))
                            _fields[r, q].ResourceType = GetNextResourceType(possibleResources);
                        else
                            _fields[r, q].ResourceType = ResourceType.Sea;

                        if (_fields[r, q].ResourceType == ResourceType.None)
                            RobberPosition = (r, q);
                    }
                }
            }
        }

        private bool IsInnerBoard(int r, int q)
        {
           return Math.Abs(q - Size / 2 + r - Size / 2) < 1 + (Size - 2) / 2 && q != 0 && q != Size - 1 && r != 0 && r != Size - 1;
        }

        private void InitializeNodes()
        {
            _nodes = new Node[Size, Size, 2];
            for (int q = 0; q < Size; ++q)
            {
                for (int r = 0; r < Size; ++r)
                {

                    if (IsInnerBoard(r, q))
                    {
                        _nodes[r, q, (int) Direction.Bottom] = new();
                        _nodes[r, q, (int) Direction.Bottom].Building = new(null);
                        _nodes[r, q, (int) Direction.Top] = new();
                        _nodes[r, q, (int) Direction.Top].Building = new(null);
                    } 
                    else if (Math.Abs(q + r - Size + 1) < 1 + Size / 2)
                    {
                        if (r > Size / 2)
                        {
                            _nodes[r, q, (int) Direction.Top] = new();
                            _nodes[r, q, (int)Direction.Top].Building = new(null);
                        }
                        else if (r < Size / 2)
                        {
                            _nodes[r, q, (int)Direction.Bottom] = new();
                            _nodes[r, q, (int)Direction.Bottom].Building = new(null);
                        }
                    }
                }
            }
        }

        private void InitializeRoads()
        {
            _roads = new List<(NodeCoordinates, byte)>[Size, Size, 2];
            for (int q = 0; q < Size; ++q)
            {
                for (int r = 0; r < Size; ++r)
                {
                    _roads[r, q, 0] = new();
                    _roads[r, q, 1] = new();
                }
            }
        }

        private void InitializePlayers(List<String> playerNames)
        {
            _currentPlayer = 0;
            _turnIncrement = 1;
            _initialTurn = true;
            _players = new Player[playerNames.Count];
            for (int i = 0; i < _players.Length; ++i)
            {
                _players[i] = new Player();
                _players[i].Name = playerNames[i];
                _players[i].Id = (byte) i;
            }
        }

        private ResourceType GetNextResourceType(Dictionary<ResourceType, int> possibleResources)
        {
            List<ResourceType> possibilitiesList = possibleResources.Keys.ToList();
            ResourceType randResource = possibilitiesList[rand.Next(possibilitiesList.Count)];
            --possibleResources[randResource];

            if (possibleResources[randResource] == 0)
                possibleResources.Remove(randResource);

            return randResource;
        }



        private List<NodeCoordinates> GetNeighborCoords(NodeCoordinates node)
        {
            List<NodeCoordinates> answer = new();
            switch (node.Dir)
            {
                case Direction.Top:
                    {
                        if (node.R > 0)
                            answer.Add(new NodeCoordinates(node.R - 1, node.Q, Direction.Bottom));
                        if (node.R > 0 && node.Q < 6)
                            answer.Add(new NodeCoordinates(node.R - 1, node.Q + 1, Direction.Bottom));
                        if (node.R > 1 && node.Q < 6)
                            answer.Add(new NodeCoordinates(node.R - 2, node.Q + 1, Direction.Bottom));
                            //return new List<NodeCoordinates> { new NodeCoordinates(node.R - 1, node.Q    , Direction.Bottom),
                            //                                   new NodeCoordinates(node.R - 1, node.Q + 1, Direction.Bottom),
                            //                                   new NodeCoordinates(node.R - 2, node.Q + 1, Direction.Bottom)};
                        return answer;
                    }
                case Direction.Bottom:
                    {
                        if (node.R < 6)
                            answer.Add(new NodeCoordinates(node.R + 1, node.Q    , Direction.Top));
                        if (node.Q > 0 && node.R < 6)
                            answer.Add(new NodeCoordinates(node.R + 1, node.Q - 1, Direction.Top));
                        if (node.R < 5 && node.Q > 0)
                            answer.Add(new NodeCoordinates(node.R + 2, node.Q - 1, Direction.Top));
                        //return new List<NodeCoordinates> { new NodeCoordinates(node.R + 1, node.Q - 1, Direction.Top),
                        //                                   new NodeCoordinates(node.R + 1, node.Q    , Direction.Top),
                        //                                   new NodeCoordinates(node.R + 2, node.Q - 1, Direction.Top)};
                        return answer;
                    }
                default:
                    throw new InvalidOperationException("Invalid Node direction!");
            }
        }
        #endregion

    }
}
