﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;
using CatanGame.ViewModel;

namespace CatanGame.View
{
    public class CatanBoardCanvas : Canvas
    {
        #region Private fields
        private CatanViewModel _viewModel;
        #endregion

        #region Public constructors
        public CatanBoardCanvas() : base()
        {
            SizeChanged += DrawMap;
        }
        #endregion

        #region Private fields


        private Ellipse[,] _roads;
        private Ellipse[,,] _settlements;
        #endregion

        #region Public Properties
        public double StrokeThickness { get; set; }
        public int Offset { get => _viewModel.BoardSize / 2; }

        public double WindowSize { get => Math.Min(ActualHeight, ActualWidth); }
        public double HexagonRadius { get => WindowSize / (_viewModel.BoardSize * 2);  }

        public CatanViewModel ViewModel 
        { 
            get => _viewModel;
            set
            {
                _viewModel = value;
                _roads = new Ellipse[_viewModel.BoardSize * _viewModel.BoardSize * 2, _viewModel.BoardSize * _viewModel.BoardSize * 2];
                _settlements = new Ellipse[_viewModel.BoardSize, _viewModel.BoardSize, 2];
                DrawMap(this, null);
            }
        }

        #endregion

        #region Private methods
        private Brush GetBrush(String fieldType)
        {
            switch (fieldType)
            {
                case "None":
                    return Brushes.SandyBrown;
                case "Lumber":
                    return Brushes.DarkOliveGreen;
                case "Brick":
                    return Brushes.SaddleBrown;
                case "Wool":
                    return Brushes.YellowGreen;
                case "Grain":
                    return Brushes.Yellow;
                case "Ore":
                    return Brushes.Gray;
                case "Sea":
                    return Brushes.LightBlue;
                default:
                    throw new InvalidOperationException("Invalid resource type!");
            }
        }

        private Brush GetPlayerColor(int playerId)
        {
            switch (playerId)
            {
                case 0:
                    return Brushes.Red;
                case 1:
                    return Brushes.Green;
                case 2:
                    return Brushes.Blue;
                default:
                    throw new InvalidOperationException("Invalid player id!");
            }
        }

        private void DrawMap(object a, SizeChangedEventArgs b)
        {
            if (_viewModel is null) return;
       
            Children.Clear();

            for (int q = 0; q < _viewModel.BoardSize; ++q)
            {
                for (int r = 0; r < _viewModel.BoardSize; ++r)
                {
                    if (_viewModel.Fields[r * _viewModel.BoardSize + q] is not null)
                    { 
                        Polygon myPolygon;

                        Point position = CalculatePositionOnCanvas(r, q);

                        myPolygon = HexagonFactory.MakeHexagon(position, HexagonRadius);
                        myPolygon.Stroke = Brushes.Black;
                        myPolygon.Fill = GetBrush(_viewModel.Fields[r * _viewModel.BoardSize + q].FieldType);
                        myPolygon.StrokeThickness = StrokeThickness;
                        //myPolygon.HorizontalAlignment = HorizontalAlignment.Left;
                        //myPolygon.VerticalAlignment = VerticalAlignment.Center;

                        Label label = new Label();
                        label.Content = q.ToString() + "," + r.ToString();
                        label.Margin = new Thickness(position.X - HexagonRadius / 2, position.Y- HexagonRadius / 2, 0, 0);

                        Children.Add(myPolygon);
                        Children.Add(label);

                        CreateRoadSpots(position, (r, q), HexagonRadius, 20);
                    }
                }
            }
            CreateBuildingSpots(20);
        }

        private void CreateRoadSpots(Point pos, (int R, int Q) coords, double hexagonRadius, double ellipseSize) 
        {
            int maxId = _viewModel.BoardSize * _viewModel.BoardSize * 2;
            double marginOffset = ellipseSize / 2;
            int[] nodeIds = new int[]
            {
                GetNodeId(coords.R - 1, coords.Q + 1, 0),
                GetNodeId(coords.R + 1, coords.Q, 1),
                GetNodeId(coords.R, coords.Q, 0),
                GetNodeId(coords.R + 1, coords.Q - 1, 1),
            };
            for (int i = 0; i < 3; ++i)
            {
                Point firstNode = HexagonFactory.HexCorner(pos, hexagonRadius, i);
                Point secondNode = HexagonFactory.HexCorner(pos, hexagonRadius, i + 1);

                Point roadPos = new Point((firstNode.X + secondNode.X) / 2, (firstNode.Y + secondNode.Y) / 2);

                Ellipse roadEllipse = new Ellipse()
                {
                    Width = ellipseSize,
                    Height = ellipseSize,
                    Margin = new Thickness(roadPos.X - marginOffset, roadPos.Y - marginOffset, 0, 0),
                    Fill = Brushes.LightGray,
                    Stroke = Brushes.Green,
                    StrokeThickness = 1,
                    Opacity = 0.5,
                    Visibility = Visibility.Hidden
                };

                roadEllipse.MouseDown += Ellipse_Click;
                SetZIndex(roadEllipse, 100);
                Children.Add(roadEllipse);
                int firstNodeId = nodeIds[i];
                int secondNodeId = nodeIds[i + 1];



                if (firstNodeId >= 0 && secondNodeId >= 0 && firstNodeId < maxId && secondNodeId < maxId)
                {
                    _roads[firstNodeId, secondNodeId] = roadEllipse;
                    _roads[secondNodeId, firstNodeId] = roadEllipse;
                    _viewModel.Roads[firstNodeId][secondNodeId].PropertyChanged +=
                        (obj, param) => UpdateRoad(firstNodeId, secondNodeId);
                    _roads[firstNodeId,secondNodeId].MouseDown += 
                        (obj, param) => _viewModel.Roads[firstNodeId][secondNodeId].BuildRoadCommand.Execute(null);
                }
            }
        }

        private void CreateBuildingSpots(double ellipseSize)
        {
            double ellipseMarginOffset = ellipseSize / 2;
            int[] cornerIds = { 2, 5 };
            foreach (NodeViewModel node in _viewModel.Nodes)
            {
                if (node is null) continue;

                (int R, int Q, int D) coords =  node.Position;
                Point position = CalculatePositionOnCanvas(coords.R, coords.Q);
                Point cornerPos = HexagonFactory.HexCorner(position, HexagonRadius, cornerIds[coords.D]);
                Ellipse ellipse = new Ellipse()
                {
                    Width = ellipseSize,
                    Height = ellipseSize,
                    Margin = new Thickness(cornerPos.X - ellipseMarginOffset, cornerPos.Y - ellipseMarginOffset, 0, 0),
                    Fill = GetBuildingBrush(node.HouseType),
                    Stroke = coords.D % 2 == 0 ? Brushes.Black : Brushes.Red,
                    StrokeThickness = 1,
                    Opacity = 0.5,
                    Visibility = Visibility.Visible,
                    
                };
                SetZIndex(ellipse, 100);
                int coordinateId = GetNodeId(coords.R, coords.Q, coords.D);
                ellipse.MouseDown += (obj, args) => node.UpgradeNodeCommand.Execute(null);
                node.PropertyChanged += (obj, param) =>
                {
                    UpdateNode(coords.R, coords.Q, coords.D);
                };
                
                Children.Add(ellipse);
                Label lab = new Label()
                {
                    Content = coordinateId,
                    Margin = new Thickness(cornerPos.X - ellipseMarginOffset, cornerPos.Y - ellipseMarginOffset, 0, 0),
                };
                SetZIndex(lab, 200);
                //Children.Add(lab);
                _settlements[coords.R, coords.Q, coords.D] = ellipse;
            }           
        }

        private Point CalculatePositionOnCanvas(int r, int q)
        {
            int offsetR = r - Offset;
            int offsetQ = q - Offset;
            double xCoord = WindowSize / 2 + HexagonRadius * (Math.Sqrt(3) * offsetQ + Math.Sqrt(3) * offsetR * 0.5);
            double yCoord = WindowSize / 2 + HexagonRadius * 1.5 * offsetR;
            return  new Point(xCoord, yCoord);
        }

        private Brush GetBuildingBrush(String buildingType)
        {
            switch(buildingType)
            {
                case "Settlement":
                    return Brushes.Black;
                case "City":
                    return Brushes.Gold;
                default:
                    return Brushes.LightGray;
            }
        }

        private void UpdateNode(int r, int q, int d)
        { 
            _settlements[r, q, d].Fill = GetPlayerColor(_viewModel.CurrentPlayerId);
            _settlements[r, q, d].Stroke = GetBuildingBrush(_viewModel.Nodes[GetNodeId(r, q, d)].HouseType);
            _settlements[r, q, d].Opacity = 1;
        }

        private void UpdateRoad(int firstNodeId, int secondNodeId)
        {
            _roads[firstNodeId, secondNodeId].Visibility = _viewModel.Roads[firstNodeId][secondNodeId].IsBuildable ? Visibility.Visible : Visibility.Hidden;
            
            if (_viewModel.Roads[firstNodeId][secondNodeId].Owner is not null)
            {
                _roads[firstNodeId, secondNodeId].Visibility = Visibility.Visible;
                _roads[firstNodeId, secondNodeId].Fill = GetPlayerColor(Byte.Parse(_viewModel.Roads[firstNodeId][secondNodeId].Owner));
                _roads[firstNodeId, secondNodeId].Opacity = 1;
            }
        }
        
        private void Ellipse_Click(object obj, MouseButtonEventArgs e)
        {
            Ellipse house = obj as Ellipse;
            Thickness t = house.Margin;
            //MessageBox.Show(((int)t.Left).ToString() + ", " + ((int)t.Top).ToString());
        }

        
        private int GetNodeId(int r, int q, int dir)
        {
            return r * _viewModel.BoardSize * 2 + q * 2 + dir;
        }

        private int GetNodeId((int R, int Q, int Dir)coords)
        {
            return GetNodeId(coords.R, coords.Q, coords.Dir);
        }
        #endregion
    }
}
